from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Exec functions for AWS API Gateway v2 Stage resources.
"""


async def update(
    hub,
    ctx,
    api_id: str,
    resource_id: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
) -> Dict[str, Any]:
    r"""
    Updates an AWS API Gateway v2 Stage resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Stage resource identifier in Amazon Web Services.
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        resource_parameters(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "AccessLogSettings": "access_log_settings",
            "AutoDeploy": "auto_deploy",
            "ClientCertificateId": "client_certificate_id",
            "DefaultRouteSettings": "default_route_settings",
            "DeploymentId": "deployment_id",
            "Description": "description",
            "RouteSettings": "route_settings",
            "StageVariables": "stage_variables",
        }
    )

    parameters_to_update = {}

    for key, value in resource_parameters.items():
        if value is not None and value != raw_resource.get(key):
            parameters_to_update[key] = resource_parameters[key]

    if parameters_to_update:
        result["ret"] = {}
        for parameter_raw, parameter_present in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

        if ctx.get("test", False):
            result["comment"] = (
                f"Would update parameters: " + ",".join(result["ret"].keys()),
            )
        else:
            update_ret = await hub.exec.boto3.client.apigatewayv2.update_stage(
                ctx,
                ApiId=api_id,
                StageName=resource_id,
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            result["comment"] = (
                f"Updated parameters: " + ",".join(result["ret"].keys()),
            )

    return result



from typing import Any
from typing import Dict

try:
    import pgpy

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


__func_alias__ = {"list_": "list"}


async def list_(
    hub, ctx, user_name: str, access_key_id: str = None, status: str = None
) -> Dict[str, Any]:
    """
    Returns a list of access keys for a user_name, optionally filtering by status and access_key_id.

    Args:
        user_name: The user_name to list keys for.
        access_key_id (Optional): match exact access key to return
        status (Optional): match "Active" or "Inactive" keys; "Active" means the key is valid for API calls.
    """
    # we manually filter status, so we must validate it
    if status not in [None, "Active", "Inactive"]:
        return {
            "result": False,
            "ret": None,
            "comment": (
                f"Could not list access keys: '{status}' isn't a valid status",
            ),
        }

    ret = await hub.exec.boto3.client.iam.list_access_keys(ctx, UserName=user_name)

    if not ret["result"]:
        return {
            "result": False,
            "ret": None,
            "comment": (f"Could not list access keys: {ret['comment']}",),
        }

    access_keys = []
    for access_key_raw in ret["ret"]["AccessKeyMetadata"]:
        tmp_access_key = (
            hub.tool.aws.iam.conversion_utils.convert_raw_access_key_to_camel_case(
                access_key_raw
            )
        )
        if access_key_id and access_key_id != tmp_access_key.get("access_key_id"):
            continue
        if status and status != tmp_access_key.get("status"):
            continue
        access_keys.append(tmp_access_key)

    result = {"result": True, "ret": access_keys, "comment": ()}
    if not len(access_keys):
        if status and access_key_id:
            result["comment"] += (
                f"List for {user_name} was successful, but no key matched {access_key_id} and {status}",
            )
        elif access_key_id:
            result["comment"] += (
                f"List for {user_name} was successful, but no key matched {access_key_id}",
            )
        elif status:
            result["comment"] += (
                f"List for {user_name} was successful, but no key matched {status}",
            )
        else:
            result["comment"] += (
                f"List for {user_name} was successful, but there were no keys.",
            )

    return result


async def update(
    hub, ctx, user_name: str, access_key_id: str, status: str
) -> Dict[str, Any]:
    """
    Updates the status of an existing access key. Does not execute if ctx["test"] is True.

    Args:
        user_name: exact user that owns the key
        access_key_id: exact access key to update
        status: "Active" or "Inactive", meaning it is valid for API calls.
    """
    result = {"result": True, "ret": None, "comment": ()}
    # this should never happen and would result in unexpected behavior
    if (not user_name) or (not access_key_id) or (not status):
        result["result"] = False
        result["comment"] = (
            "user_name, access_key_id and status all need to be supplied",
        )
        return result

    ret = await hub.exec.boto3.client.iam.update_access_key(
        ctx, AccessKeyId=access_key_id, Status=status, UserName=user_name
    )

    if not ret["result"]:
        result["result"] = False
        result["comment"] += (
            f"Could not update access key {access_key_id} for {user_name}",
        )
        result["comment"] += ret["comment"]
    else:
        result["comment"] += (
            f"Access Key {access_key_id} for {user_name} set to '{status}'",
        )
    return result


async def delete(hub, ctx, user_name: str, access_key_id: str) -> Dict[str, Any]:
    """
    Deletes an existing access key, does not execute if ctx["test"] is set.

    "result" is True if delete succeeded or no such key or user exists.
    "result" is False if there is an error deleting the key

    Args:
        user_name: exact user owning the key
        access_key_id: exact access key to delete
    """
    result = {"result": True, "ret": None, "comment": ()}
    # this should never happen and would result in unexpected behavior
    assert user_name and access_key_id

    ret = await hub.exec.boto3.client.iam.delete_access_key(
        ctx, AccessKeyId=access_key_id, UserName=user_name
    )

    if not ret["result"]:
        if "NoSuchEntityException" in str(ret["comment"]):
            result["comment"] = (
                f"Access key {access_key_id} for {user_name} does not exist",
            )
        else:
            result["result"] = False
            result["comment"] = (
                f"Could not delete access key '{access_key_id}'",
            ) + ret["comment"]
    else:
        result["comment"] = (f"Access key '{access_key_id}' deleted",)
    return result


async def create(hub, ctx, user_name: str, pgp_key: str = None) -> Dict[str, Any]:
    """
    Creates a new access key for the specified user.

    Args:
        user_name: exact user the key will be created for.
        pgp_key (Optional): base 64 encoded public pgp key with which we will encrypt the returned secret_access_key.

    """

    result = {"result": True, "ret": None, "comment": ()}

    ret = await hub.exec.boto3.client.iam.create_access_key(ctx, UserName=user_name)

    if not ret["result"]:
        result["result"] = False
        result["comment"] += (f"Could not create access key: {ret['comment']}",)
    else:
        access_key = (
            hub.tool.aws.iam.conversion_utils.convert_raw_access_key_to_camel_case(
                ret["ret"]["AccessKey"]
            )
        )
        if pgp_key:
            decoded_pgp_key = hub.tool.aws.b64.decode(pgp_key)
            loaded_pgp_key, _ = pgpy.PGPKey.from_blob(decoded_pgp_key)
            secret_access_key = pgpy.PGPMessage.new(
                access_key["secret_access_key"],
                compression=pgpy.constants.CompressionAlgorithm.Uncompressed,
            )
            loaded_pgp_key.encrypt(secret_access_key)
            access_key["secret_access_key"] = hub.tool.aws.b64.encode(secret_access_key)

        result["ret"] = access_key
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.eks.create_fargate_profile
hub.exec.boto3.client.eks.delete_fargate_profile
hub.exec.boto3.client.eks.describe_fargate_profile
hub.exec.boto3.client.eks.list_fargate_profiles
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


TREQ = {
    "present": {
        "require": [
            "aws.eks.cluster.present",
        ],
    },
}
resource_type = "aws.eks.fargate_profile"


async def present(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    pod_execution_role_arn: str,
    resource_id: str = None,
    subnets: List = None,
    selectors: List = None,
    client_request_token: str = None,
    tags: Dict = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""

    Creates an Fargate profile for your Amazon EKS cluster. You must have at least one Fargate profile in a cluster
    to be able to run pods on Fargate. The Fargate profile allows an administrator to declare which pods run on
    Fargate and specify which pods run on which Fargate profile. This declaration is done through the profile’s
    selectors. Each profile can have up to five selectors that contain a namespace and labels. A namespace is
    required for every selector. The label field consists of multiple optional key-value pairs. Pods that match the
    selectors are scheduled on Fargate. If a to-be-scheduled pod matches any of the selectors in the Fargate
    profile, then that pod is run on Fargate. When you create a Fargate profile, you must specify a pod execution
    role to use with the pods that are scheduled with the profile. For more information, see Pod Execution Role in the
    Amazon EKS User Guide. Fargate profiles are immutable. However, you can create a new updated profile to replace an existing profile
    and then delete the original after the updated profile has finished creating. If any Fargate profiles in a cluster are in
    the DELETING status, you must wait for that Fargate profile to finish deleting before you can create any other
    profiles in that cluster. For more information, see Fargate Profile in the Amazon EKS User Guide.

    Args:
        name(Text): An Idem name of the resource
        cluster_name(Text): The name of the Amazon EKS cluster to apply the Fargate profile to.
        pod_execution_role_arn(Text): The Amazon Resource Name (ARN) of the pod execution role to use for pods that match the
            selectors in the Fargate profile.
        subnets(List, optional): The IDs of private subnets to launch your pods into.
        Private subnets selected for fargate profile creation must be in different availability zones. Defaults to None.
        selectors(List, optional): The selectors to match for pods to use this Fargate profile. Each selector must have an
            associated namespace. Optionally, you can also specify labels for a namespace. You may specify
            up to five selectors in a Fargate profile. Defaults to None.
        client_request_token(Text, optional): Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. Defaults to None.
        tags(Dict, optional): The metadata to apply to the Fargate profile to assist with categorization and organization.
            Each tag consists of a key and an optional value. Defaults to None.
        resource_id(Text, optional): AWS Fargate Profile name
        timeout(Dict, optional): Timeout configuration for operations on fargate_profile
            * create (Dict) -- Timeout configuration for creating fargate_profile
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- The maximum number of attempts to be made.

        Request Syntax:
            fargate_resource_id:
               aws.eks.fargate_profile.present:
               - name: 'string'
               - resource_id: 'string'
               - cluster_name: 'string'
               - pod_execution_role_arn: 'string'
               - subnets:
                  - private_subnet_id
                  - private_subnet_id
            - selectors:
                   - labels: {}
                     namespace: 'string'


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            fargate01-idem-cluster01:
              aws.eks.fargate_profile.present:
              - name: fargate01
              - cluster_name: idem-cluster01
              - pod_execution_role_arn: arn:aws:iam::537227425989:role/eksfargateprofile-role
              - subnets:
                - subnet-006a41c410b9485dd
                - subnet-01368936bb540ba1e
              - selectors:
                - labels: {}
                  namespace: default
              - tags:
                  fargate01: 'created via code'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.eks.describe_fargate_profile(
        ctx,
        clusterName=cluster_name,
        fargateProfileName=resource_id if resource_id else name,
    )

    # existing resource
    if before and before["ret"]["fargateProfile"]:
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
            raw_resource=before["ret"]["fargateProfile"]
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (f"aws.eks.fargate_profile '{name}' already exists",)
        # fargate_profile can only be updated for tags after its creation, but tag update apis are currently
        # supported for eks clusters and nodegroups only. So no update operations are performed
        return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cluster_name": cluster_name,
                    "resource_id": resource_id,
                    "client_request_token": client_request_token,
                    "pod_execution_role_arn": pod_execution_role_arn,
                    "tags": tags,
                    "subnets": subnets,
                    "selectors": selectors,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=resource_type, name=name
            )
            return result
        # create a new resource
        ret = await hub.exec.boto3.client.eks.create_fargate_profile(
            ctx,
            fargateProfileName=name,
            clusterName=cluster_name,
            podExecutionRoleArn=pod_execution_role_arn,
            subnets=subnets,
            selectors=selectors,
            clientRequestToken=client_request_token,
            tags=tags,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        # create waiter config
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=10,
            default_max_attempts=60,
            timeout_config=timeout.get("create") if timeout else None,
        )
        hub.log.debug(f"Waiting on creating aws.eks.fargate_profile '{name}'")
        try:
            # wait till fargate_profile is active,
            await hub.tool.boto3.client.wait(
                ctx,
                "eks",
                "fargate_profile_active",
                None,
                clusterName=cluster_name,
                fargateProfileName=name,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type=resource_type, name=name
        )

    # Get the newly created resource
    result[
        "new_state"
    ] = hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
        raw_resource=ret["ret"]["fargateProfile"]
    )
    return result


async def absent(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""

    Deletes an Fargate profile. When you delete a Fargate profile, any pods running on Fargate that were created
    with the profile are deleted. If those pods match another Fargate profile, then they are scheduled on Fargate
    with that profile. If they no longer match any Fargate profiles, then they are not scheduled on Fargate and they
    may remain in a pending state. Only one Fargate profile in a cluster can be in the DELETING status at a time.
    You must wait for a Fargate profile to finish deleting before you can delete any other profiles in that cluster.

    Args:
        name(Text): An Idem name of the resource
        cluster_name(Text): The name of the Amazon EKS cluster associated with the Fargate profile to delete.
        resource_id(Text, Optional): AWS Fargate profile name. Idem considers name of fargate profile if
         this field is not specified.
        timeout(Dict, optional): Timeout configuration for operations on fargate_profile
            * delete (Dict) -- Timeout configuration for deleting fargate_profile
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- The maximum number of attempts to be made

    Request Syntax:
        fargate_resource_id:
           aws.eks.fargate_profile.absent:
           - name: 'string'
           - cluster_name: 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls
            fargate01-idem-cluster01:
              aws.eks.fargate_profile.absent:
              - resource_id: fargate01
              - name: fargate01
              - cluster_name: idem-cluster01
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.eks.describe_fargate_profile(
        ctx,
        clusterName=cluster_name,
        fargateProfileName=resource_id if resource_id else name,
    )
    if before and before["ret"]["fargateProfile"]:
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
            raw_resource=before["ret"]["fargateProfile"]
        )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=resource_type, name=name
        )
    else:
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type=resource_type, name=name
            )
        else:
            ret = await hub.exec.boto3.client.eks.delete_fargate_profile(
                ctx,
                clusterName=cluster_name,
                fargateProfileName=resource_id if resource_id else name,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            # wait till fargate_profile has been deleted
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=30,
                default_max_attempts=60,
                timeout_config=timeout.get("delete") if timeout else None,
            )
            hub.log.debug(f"Waiting on deleting aws.eks.fargate_profile '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "eks",
                    "fargate_profile_deleted",
                    None,
                    fargateProfileName=resource_id,
                    clusterName=cluster_name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type=resource_type, name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function
    Lists the Fargate profiles associated with the specified cluster in your Amazon Web Services account in the
    specified Region.
    Returns:
        Dict[str, Dict[str, Any]]
    Examples:

        .. code-block:: sls
         fargate01-idem-cluster01:
              aws.eks.fargate_profile.present:
              - name: fargate01
              - cluster_name: idem-cluster01
              - pod_execution_role_arn: arn:aws:iam::537227425989:role/eksfargateprofile-role
              - subnets:
                - subnet-006a41c410b9485dd
                - subnet-01368936bb540ba1e
              - selectors:
                - labels: {}
                  namespace: default
              - tags:
                  fargate01: 'created via code'
    """

    result = {}
    # List all clusters
    cluster_ret = await hub.exec.boto3.client.eks.list_clusters(ctx)
    if not cluster_ret["result"]:
        hub.log.debug(f"Could not list clusters {cluster_ret['comment']}")
        return {}

    fargate_profile_list = []
    # Get fargate profiles in each cluster
    for clusterName in cluster_ret["ret"]["clusters"]:
        fargate_profiles_ret = await hub.exec.boto3.client.eks.list_fargate_profiles(
            ctx, clusterName=clusterName
        )
        if not fargate_profiles_ret["result"]:
            hub.log.info(
                f"No fargate profiles could be listed for cluster {clusterName}. Fetching the next cluster."
            )
            continue

        # maintain a list of clusterName and fargateProfileName for future use
        for fargate_profile_name in fargate_profiles_ret["ret"]["fargateProfileNames"]:
            fargate_profile_list.append(
                {
                    "clusterName": clusterName,
                    "fargateProfileName": fargate_profile_name,
                }
            )
    # Describe each fargate profile
    for fargate_profile in fargate_profile_list:
        clusterName = fargate_profile["clusterName"]
        fargateProfileName = fargate_profile["fargateProfileName"]
        describe_ret = await hub.exec.boto3.client.eks.describe_fargate_profile(
            ctx, clusterName=clusterName, fargateProfileName=fargateProfileName
        )

        if not describe_ret["result"]:
            hub.log.info(
                f"Describe command failed for resource {fargateProfileName} under cluster {clusterName}"
            )
            continue

        fargate_profile = describe_ret["ret"]["fargateProfile"]
        translated_resource = (
            hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
                raw_resource=fargate_profile
            )
        )
        # state_id is combination of fargateProfileName and clusterName to maintain integrity as more than one
        # cluster can have same fargate profile names
        state_id = (
            fargate_profile.get("fargateProfileName")
            + "-"
            + fargate_profile.get("clusterName")
        )
        result[state_id] = {
            "aws.eks.fargate_profile.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.apigatewayv2.create_domain_name
hub.exec.boto3.client.apigatewayv2.delete_domain_name
hub.exec.boto3.client.apigatewayv2.get_domain_name
hub.exec.boto3.client.apigatewayv2.get_domain_names
hub.exec.boto3.client.apigatewayv2.tag_resource
hub.exec.boto3.client.apigatewayv2.untag_resource
hub.exec.boto3.client.apigatewayv2.update_domain_name
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    domain_name_configurations: List[
        make_dataclass(
            "DomainNameConfiguration",
            [
                ("ApiGatewayDomainName", str, field(default=None)),
                ("CertificateArn", str, field(default=None)),
                ("CertificateName", str, field(default=None)),
                ("CertificateUploadDate", str, field(default=None)),
                ("DomainNameStatus", str, field(default=None)),
                ("DomainNameStatusMessage", str, field(default=None)),
                ("EndpointType", str, field(default=None)),
                ("HostedZoneId", str, field(default=None)),
                ("SecurityPolicy", str, field(default=None)),
                ("OwnershipVerificationCertificateArn", str, field(default=None)),
            ],
        )
    ] = None,
    mutual_tls_authentication: make_dataclass(
        "MutualTlsAuthenticationInput",
        [
            ("TruststoreUri", str, field(default=None)),
            ("TruststoreVersion", str, field(default=None)),
        ],
    ) = None,
    tags: Dict = None,
) -> Dict[str, Any]:
    r"""
    Creates an API Gateway v2 domain name resource.

    Args:
        name(string): An Idem name of the resource. This is also used as the name of the domain name during resource creation.
        resource_id(string, optional): The domain name resource identifier in Amazon Web Services.
        domain_name_configurations(List, optional): The domain name configurations. Defaults to None.
            * (ApiGatewayDomainName, optional): A domain name for the API.
            * (CertificateArn, optional): An AWS-managed certificate that will be used by the edge-optimized endpoint for this domain
                name. AWS Certificate Manager is the only supported source.
            * (CertificateName, optional): The user-friendly name of the certificate that will be used by the edge-optimized endpoint for
                this domain name.
            * (CertificateUploadDate, optional): The timestamp when the certificate that was used by edge-optimized endpoint for this domain name
                was uploaded.
            * (DomainNameStatus, optional): The status of the domain name migration. The valid values are AVAILABLE, UPDATING,
                PENDING_CERTIFICATE_REIMPORT, and PENDING_OWNERSHIP_VERIFICATION. If the status is UPDATING, the
                domain cannot be modified further until the existing operation is complete. If it is AVAILABLE,
                the domain can be updated.
            * (DomainNameStatusMessage, optional): An optional text message containing detailed information about status of the domain name
            migration.
            * (EndpointType, optional): The endpoint type.
            * (HostedZoneId, optional): The Amazon Route 53 Hosted Zone ID of the endpoint.
            * (SecurityPolicy, optional): The Transport Layer Security (TLS) version of the security policy for this domain name. The
            valid values are TLS_1_0 and TLS_1_2.
            * (OwnershipVerificationCertificateArn, optional): The ARN of the public certificate issued by ACM to validate ownership of your custom domain.
                Only required when configuring mutual TLS and using an ACM imported or private CA certificate
                ARN as the regionalCertificateArn
        mutual_tls_authentication(Dict, optional): The mutual TLS authentication configuration for a custom domain name. Defaults to None.
            * (TruststoreUri, optional): An Amazon S3 URL that specifies the truststore for mutual TLS authentication, for example,
                s3://bucket-name/key-name. The truststore can contain certificates from public or private
                certificate authorities. To update the truststore, upload a new version to S3, and then update
                your custom domain name to use the new version. To update the truststore, you must have
                permissions to access the S3 object.
            * (TruststoreVersion, optional): The version of the S3 object that contains your truststore. To specify a version, you must have
                versioning enabled for the S3 bucket.
        tags(Dict, optional): The collection of tags associated with a domain name.

    Request Syntax:
        [idem_test_aws_apigatewayv2_domain_name]:
          aws.apigatewayv2.domain_name.present:
            - name: 'string'
            - domain_name_configurations: [
                {
                    'ApiGatewayDomainName': 'string',
                    'CertificateArn': 'string',
                    'CertificateName': 'string',
                    'CertificateUploadDate': datetime(2015, 1, 1),
                    'DomainNameStatus': 'AVAILABLE'|'UPDATING'|'PENDING_CERTIFICATE_REIMPORT'|'PENDING_OWNERSHIP_VERIFICATION',
                    'DomainNameStatusMessage': 'string',
                    'EndpointType': 'REGIONAL'|'EDGE',
                    'HostedZoneId': 'string',
                    'SecurityPolicy': 'TLS_1_0'|'TLS_1_2',
                    'OwnershipVerificationCertificateArn': 'string'
                }
            ]
            - mutual_tls_authentication: {
                'TruststoreUri': 'string',
                'TruststoreVersion': 'string
              }
            - tags: {
                'string': 'string'
              }

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_domain_name:
              aws.apigatewayv2.domain_name.present:
                - name: value
                - domain_name_configurations: [
                    {
                        'ApiGatewayDomainName': value,
                        'CertificateArn': 'value,
                        'CertificateName': 'value,
                        'CertificateUploadDate': value,
                        'DomainNameStatus': value,
                        'DomainNameStatusMessage': value,
                        'EndpointType': value,
                        'HostedZoneId': value,
                        'SecurityPolicy': value,
                        'OwnershipVerificationCertificateArn': value
                    }
                ]
                - mutual_tls_authentication: {
                    'TruststoreUri': 'string',
                    'TruststoreVersion': 'string
                }
                - tags: {
                    value: value
                }
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None
    resource_updated = False

    if resource_id:
        before_ret = await hub.exec.boto3.client.apigatewayv2.get_domain_name(
            ctx, DomainName=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigatewayv2.domain_name", name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.apigatewayv2.domain_name.convert_raw_domain_name_to_present(
            raw_resource=before_ret.get("ret")
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "DomainNameConfigurations": domain_name_configurations,
            "MutualTlsAuthentication": mutual_tls_authentication,
        }

        update_domain_name_ret = await hub.exec.aws.apigatewayv2.domain_name.update(
            ctx,
            resource_id=resource_id,
            raw_resource=before_ret.get("ret"),
            resource_parameters=resource_parameters,
        )
        result["comment"] = result["comment"] + update_domain_name_ret["comment"]
        if not update_domain_name_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_domain_name_ret["ret"])
        if update_domain_name_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_domain_name_ret["ret"])

        if tags is not None and tags != result["old_state"].get("tags"):
            resource_arn = hub.tool.aws.arn_utils.build_arn(
                partition="aws",
                service="apigateway",
                region=ctx["acct"]["region_name"],
                resource="/domainnames/" + resource_id,
            )
            update_tags_ret = await hub.exec.aws.apigatewayv2.tag.update_tags(
                ctx,
                resource_arn=resource_arn,
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            if not update_tags_ret["result"]:
                result["result"] = False
                return result

            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if update_tags_ret["ret"] and ctx.get("test", False):
                result["new_state"]["tags"] = update_tags_ret["ret"].get("tags")

        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "domain_name_configurations": domain_name_configurations,
                    "name": name,
                    "mutual_tls_authentication": mutual_tls_authentication,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigatewayv2.domain_name", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.apigatewayv2.create_domain_name(
            ctx,
            DomainName=name,
            DomainNameConfigurations=domain_name_configurations,
            MutualTlsAuthentication=mutual_tls_authentication,
            Tags=tags,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigatewayv2.domain_name", name=name
        )
        resource_id = create_ret["ret"]["DomainName"]

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.boto3.client.apigatewayv2.get_domain_name(
            ctx, DomainName=resource_id
        )
        if not after_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + after_ret["comment"]
            return result

        result[
            "new_state"
        ] = hub.tool.aws.apigatewayv2.domain_name.convert_raw_domain_name_to_present(
            raw_resource=after_ret.get("ret")
        )

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    Deletes an API Gateway v2 domain name resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The domain name resource identifier in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

    Request syntax:
        [idem_test_aws_apigatewayv2_domain_name]:
          aws.apigatewayv2.domain_name.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_domain_name:
              aws.apigatewayv2.domain_name.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.domain_name", name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.apigatewayv2.get_domain_name(
        ctx, DomainName=resource_id
    )
    if not before_ret["result"]:
        if "NotFoundException" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.apigatewayv2.domain_name", name=name
            )
        return result

    result[
        "old_state"
    ] = hub.tool.aws.apigatewayv2.domain_name.convert_raw_domain_name_to_present(
        raw_resource=before_ret["ret"],
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigatewayv2.domain_name", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.apigatewayv2.delete_domain_name(
        ctx, DomainName=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.apigatewayv2.domain_name", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the API Gateway v2 domain name resources for an AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigatewayv2.domain_name
    """

    result = {}

    describe_ret = await hub.exec.boto3.client.apigatewayv2.get_domain_names(ctx)
    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe domain names {describe_ret['comment']}")
        return result

    for domain_name in describe_ret["ret"]["Items"]:
        resource_translated = (
            hub.tool.aws.apigatewayv2.domain_name.convert_raw_domain_name_to_present(
                raw_resource=domain_name,
            )
        )

        result[resource_translated["name"]] = {
            "aws.apigatewayv2.domain_name.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result


async def search(hub, ctx, name, resource_id: str) -> Dict[str, Any]:
    r"""
    Use an un-managed API Gateway v2 domain_name resource as a data-source.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string): The domain name resource identifier in Amazon Web Services.

    Request syntax:
        [idem_test_aws_apigatewayv2_domain_name]:
          aws.apigatewayv2.domain_name.search:
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_domain_name:
              aws.apigatewayv2.domain_name.search:
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    search_ret = await hub.exec.boto3.client.apigatewayv2.get_domain_name(
        ctx, DomainName=resource_id
    )
    if not search_ret["result"]:
        result["result"] = False
        if "NotFoundException" not in str(search_ret["comment"][0]):
            result["comment"] = search_ret["result"]
        else:
            result["comment"] = (
                f"Unable to find aws.apigatewayv2.domain_name resource with resource id '{resource_id}'",
            )
        return result

    result[
        "old_state"
    ] = hub.tool.aws.apigatewayv2.domain_name.convert_raw_domain_name_to_present(
        raw_resource=search_ret["ret"],
    )
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result



from typing import Any
from typing import Dict

from dict_tools import differ


def get_resource_vpc_config_changes(
    hub, old_vpc_config: Dict, new_vpc_config: Dict
) -> Dict[str, Any]:
    """
    Returns updated resource vpc config changes

     Args:
        hub: required for functions in hub
        old_vpc_config(Dict): old resource vpc config changes
        new_vpc_config(Dict): new resource vpc config changes

    Returns: A dict with resource vpc config changes
    """
    final_vpc_config = {}
    if new_vpc_config.get(
        "endpointPrivateAccess"
    ) is not None and not old_vpc_config.get(
        "endpointPrivateAccess"
    ) == new_vpc_config.get(
        "endpointPrivateAccess"
    ):
        final_vpc_config["endpointPrivateAccess"] = new_vpc_config.get(
            "endpointPrivateAccess"
        )

    if new_vpc_config.get(
        "endpointPublicAccess"
    ) is not None and not old_vpc_config.get(
        "endpointPublicAccess"
    ) == new_vpc_config.get(
        "endpointPublicAccess"
    ):
        final_vpc_config["endpointPublicAccess"] = new_vpc_config.get(
            "endpointPublicAccess"
        )

    if new_vpc_config.get("publicAccessCidrs") and not old_vpc_config.get(
        "publicAccessCidrs"
    ) == new_vpc_config.get("publicAccessCidrs"):
        final_vpc_config["publicAccessCidrs"] = new_vpc_config.get("publicAccessCidrs")
    return final_vpc_config


def update_labels(old_labels: dict, new_labels: dict) -> Dict[str, Any]:
    """
    Returns updated labels used in update node group config

     Args:
        old_labels(Dict): old labels
        new_labels(Dict): new labels

    Returns: A dict with labels
    """
    labels_to_add = {}
    labels_to_remove = []
    for key, value in new_labels.items():
        if (key in old_labels and old_labels.get(key) != new_labels.get(key)) or (
            key not in old_labels
        ):
            labels_to_add[key] = value

    for key in old_labels:
        if key not in new_labels:
            labels_to_remove.append(key)

    final_labels = {}
    if labels_to_add:
        final_labels["addOrUpdateLabels"] = labels_to_add
    if labels_to_remove:
        final_labels["removeLabels"] = labels_to_remove
    return final_labels


def update_taints(old_taints: list, new_taints: list) -> Dict[str, Any]:
    """
    Returns updated taints used in update node group config

     Args:
        old_taints(list): old taints
        new_taints(list): new taints

    Returns: A dict with taints
    """
    taints_to_add = list()
    taints_to_remove = list()
    old_taints_map = {taint.get("key"): taint for taint in old_taints or []}
    if new_taints is not None:
        for tag in new_taints:
            if tag.get("key") in old_taints_map:
                if tag.get("value") != old_taints_map.get(tag.get("key")).get("value"):
                    taints_to_add.append(tag)
                del old_taints_map[tag.get("key")]
            else:
                taints_to_add.append(tag)
        taints_to_remove = [taint for taint in old_taints_map.values()]
    final_taints = {}
    if taints_to_add:
        final_taints["addOrUpdateTaints"] = taints_to_add
    if taints_to_remove:
        final_taints["removeTaints"] = taints_to_remove

    return final_taints


def get_updated_node_group_config(
    hub,
    old_node_group_config: Dict,
    labels: dict,
    taints: list,
    scaling_config: Dict,
    update_config: Dict,
) -> Dict[str, Any]:
    """
    Returns updated node group config

     Args:
        hub: required for functions in hub
        old_node_group_config(Dict): old node group config
        labels(dict): new labels
        taints(List): new taints
        scaling_config(Dict): new scaling config
        update_config(Dict): new update config

    Returns: A dict with node group config
    """
    final_node_group_config = {}
    current_labels = update_labels(old_node_group_config.get("labels"), labels)
    if current_labels:
        final_node_group_config["labels"] = current_labels
    current_taints = update_taints(old_node_group_config.get("taints"), taints)
    if current_taints:
        final_node_group_config["taints"] = current_taints
    if differ.deep_diff(old_node_group_config.get("scaling_config"), scaling_config):
        final_node_group_config["scaling_config"] = scaling_config
    if differ.deep_diff(old_node_group_config.get("update_config"), update_config):
        final_node_group_config["update_config"] = update_config
    return final_node_group_config



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_policy(
    hub,
    ctx,
    policy_arn: str,
    policy_version_id: str,
    new_policy_document: str,
    timeout: Dict = None,
) -> Dict[str, Any]:
    """
    Update the policy document of an existing managed policy.

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        policy_arn: The Amazon Resource Name (ARN) of the managed policy that you want information about.
        policy_version_id: Identifies the policy version to retrieve. This parameter allows (through its regex pattern)
         a string of characters that consists of the lowercase letter 'v' followed by one or two digits, and optionally
          followed by a period '.' and a string of letters and digits.
        new_policy_document: The JSON policy document that you want to use as the content for the new policy.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS IAM Policy.
            * update (string) -- Timeout configuration for updating AWS IAM Policy
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Returns:
        {"result": True|False, "comment": Tuple, "ret": None}
    """
    result = dict(comment=(), result=True, ret=None)
    if new_policy_document is None:
        result["comment"] = (
            f"Skip aws.iam.policy update. New policy document is None",
        )
        return result
    get_ret = await hub.exec.boto3.client.iam.get_policy_version(
        ctx=ctx, PolicyArn=policy_arn, VersionId=policy_version_id
    )
    if not get_ret["result"]:
        result["result"] = False
        result["comment"] = get_ret["comment"]
        return result
    old_policy = get_ret["ret"]
    if old_policy.get("PolicyVersion").get("Document") == new_policy_document:
        result["comment"] = ("No need to update aws.iam.policy policy document",)
        return result
    try:
        # Update the policy document by creating a new policy document, set it to default, and then delete the old
        # policy document
        create_ret = await hub.exec.boto3.client.iam.create_policy_version(
            ctx=ctx,
            PolicyArn=policy_arn,
            PolicyDocument=new_policy_document,
            SetAsDefault=True,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=1,
            default_max_attempts=40,
            timeout_config=timeout.get("update") if timeout else None,
        )
        hub.log.debug(f"Waiting on updating aws.iam.policy '{policy_arn}'")
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "iam",
                "policy_exists",
                PolicyArn=policy_arn,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
        delete_ret = await hub.exec.boto3.client.iam.delete_policy_version(
            ctx=ctx, PolicyArn=policy_arn, VersionId=policy_version_id
        )
        if not delete_ret["result"]:
            hub.log.warning(
                f"hub.exec.aws.iam.policy.update_policy failed to delete policy document with version"
                f" {policy_version_id} . Manual cleanup may be needed."
            )
            result["comment"] = (
                f"hub.exec.aws.iam.policy.update_policy failed to delete policy document with version"
                f" {policy_version_id} . Manual cleanup may be needed.",
            )
        else:
            result["comment"] = (
                f"hub.exec.aws.iam.policy.update_policy updated policy document to version"
                f" {create_ret['ret'].get('PolicyVersion').get('VersionId')}",
            )
    except Exception as e:
        result["comment"] = (str(e),)
        result["result"] = False
    return result


async def update_policy_tags(
    hub,
    ctx,
    police_arn: str,
    old_tags: List[Dict[str, Any]] or Dict[str, str] = None,
    new_tags: List[Dict[str, Any]] or Dict[str, str] = None,
) -> Dict[str, Any]:
    """
    Update tags of AWS IAM Policy

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        police_arn: The ARN of the IAM customer managed policy to which you want to add tags.
         This parameter allows (through its regex pattern) a string of characters consisting of upper and lowercase
          alphanumeric characters with no spaces. You can also include any of the following characters: _+=,.@-
        old_tags (List or Dict): list of old tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                                 {tag-key: tag-value}
        new_tags (List or Dict): list of new tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                                 {tag-key: tag-value}

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}
    """
    tags_to_add = {}
    tags_to_remove = {}
    if isinstance(old_tags, List):
        old_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(old_tags)
    if isinstance(new_tags, List):
        new_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(new_tags)
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.iam.untag_policy(
                ctx, PolicyArn=police_arn, TagKeys=list(tags_to_remove.keys())
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_policy(
                ctx,
                PolicyArn=police_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    result["comment"] = (
        f"Update policy with arn '{police_arn}' tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",
    )
    return result



from typing import List


async def update_routes(
    hub,
    ctx,
    route_table_id: str,
    old_routes: List = None,
    new_routes: List = None,
):
    """
    Update routes of a route table. This function compares the existing(old) routes of route table with new routes.
    routes that are in the new route table but not in the old
    route table will be associated to route table. routes that are in the
    old route table  but not in the new route table will be disassociated from transit route table.

    Args:
        hub:
        ctx:
        route_table_id(Text): The AWS resource id of the existing route table
        old_routes(List):  routes of existing route table
        new_routes(List): new routes to update

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)

    # compare old_routes if routes are modified
    if new_routes is []:
        result["result"] = False
        result["comment"] = (
            "Route Table routes cannot be None. There should be at least one route in Route Table",
        )
        return result
    elif new_routes is not None:
        routes_to_modify = (
            hub.tool.aws.ec2.route_table_utils.get_route_table_routes_modifications(
                old_routes, new_routes
            )
        )
        if not routes_to_modify["result"]:
            result["comment"] = routes_to_modify["comment"]
            result["result"] = False
            return result
        routes_to_add = routes_to_modify["routes_to_add"]
        routes_to_delete = routes_to_modify["routes_to_delete"]
        routes_to_replace = routes_to_modify["routes_to_replace"]

        if not ctx.get("test", False):
            if routes_to_delete:
                for route_to_delete in routes_to_delete:
                    ret = await hub.exec.boto3.client.ec2.delete_route(
                        ctx, RouteTableId=route_table_id, **route_to_delete
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Deleted Routes: {routes_to_delete}, ",
                )
            if routes_to_replace:
                for route_to_replace in routes_to_replace:
                    ret = await hub.exec.boto3.client.ec2.replace_route(
                        ctx, RouteTableId=route_table_id, **route_to_replace
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Replace Routes: {routes_to_replace}, ",
                )
            if routes_to_add:
                for route_to_add in routes_to_add:
                    ret = await hub.exec.boto3.client.ec2.create_route(
                        ctx, RouteTableId=route_table_id, **route_to_add
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Added Routes: {routes_to_add}, ",
                )
    if result["comment"]:
        result["comment"] = result["comment"] + (
            f"Updated route table {route_table_id}",
        )
    result["ret"] = {
        "routes": new_routes,
    }
    return result



from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource", "soft_fail"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.vpc.present",
            "aws.ec2.subnet.present",
            "aws.iam.role.present",
        ],
    },
}

"""
run_instances options that aren't yet in the present state params:

    ipv6AddressCount - is this managed from "network_interfaces"?
    Ipv6Addressses - is this managed from "network_interfaces"?
"""


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    # From DescribeInstances
    image_id: str = None,
    instance_type: str = None,
    block_device_mappings: List[
        make_dataclass(
            "BlockDeviceMapping",
            [
                ("DeviceName", str, field(default=None)),
                ("VirtualName", str, field(default=None)),
                ("Ebs", Dict, field(default=None)),
                ("NoDevice", str, field(default=None)),
            ],
        )
    ] = None,
    ebs_optimized: bool = None,
    kernel_id: str = None,
    placement: Dict[str, Any] = None,
    subnet_id: str = None,
    network_interfaces: List[
        make_dataclass(
            "InstanceNetworkInterfaceSpecification",
            [
                ("AssociatePublicIpAddress", bool, field(default=None)),
                ("DeleteOnTermination", bool, field(default=None)),
                ("Description", str, field(default=None)),
                ("DeviceIndex", int, field(default=None)),
                ("Groups", List, field(default=None)),
                ("Ipv6AddressCount", int, field(default=None)),
                ("Ipv6Addresses", List, field(default=None)),
                ("NetworkInterfaceId", str, field(default=None)),
                ("PrivateIpAddress", str, field(default=None)),
                ("PrivateIpAddresses", List, field(default=None)),
                ("SecondaryPrivateIpAddressCount", int, field(default=None)),
                ("SubnetId", str, field(default=None)),
                ("AssociateCarrierIpAddress", bool, field(default=None)),
                ("InterfaceType", str, field(default=None)),
                ("NetworkCardIndex", int, field(default=None)),
                ("Ipv4Prefixes", List, field(default=None)),
                ("Ipv4PrefixCount", int, field(default=None)),
                ("Ipv6Prefixes", List, field(default=None)),
                ("Ipv6PrefixCount", int, field(default=None)),
            ],
        )
    ] = None,
    monitoring_enabled: bool = None,
    root_device_name: str = None,
    product_codes: List[Dict[str, str]] = None,
    source_dest_check: bool = None,
    running: bool = None,
    private_ip_address: str = None,
    reservation_id: str = None,
    owner_id: str = None,
    # From launchTemplate
    user_data: str = None,
    disable_api_termination: bool = None,
    instance_initiated_shutdown_behavior: str = None,
    enclave_options_enabled: bool = None,
    ram_disk_id: str = None,
    tags: Dict[str, str] = None,
    elastic_gpu_specifications: List[
        make_dataclass("ElasticGpuSpecification", [("Type", str, field(default=None))])
    ] = None,
    elastic_inference_accelerators: List[
        make_dataclass(
            "ElasticInferenceAccelerator",
            [("Type", str, field(default=None)), ("Count", int, field(default=None))],
        )
    ] = None,
    iam_instance_profile: Dict[str, str] = None,
    key_name: str = None,
    # Can only be changed on initial creation of an instance as far as we know so far
    instance_market_options: Dict[str, Any] = None,
    credit_specification: Dict[str, Any] = None,
    cpu_options: Dict[str, Any] = None,
    capacity_reservation_specification: Dict[str, str] = None,
    license_specifications: List[
        make_dataclass(
            "LicenseConfigurationRequest",
            [("LicenseConfigurationArn", str, field(default=None))],
        )
    ] = None,
    hibernation_enabled: bool = None,
    metadata_options: Dict[str, Any] = None,
    instance_requirements: Dict[str, Any] = None,
    private_dns_name_options: Dict[str, Any] = None,
    maintenance_options: Dict[str, Any] = None,
    sriov_net_support: str = None,
    # idem-heist options
    bootstrap: List[Dict[str, str]] = None,
    **kwargs,
) -> Dict[str, Any]:
    """
    Launches an instance using an AMI for which you have permissions.
    You can specify a number of options, or leave the default options.
    The following rules apply:
        - [EC2-VPC] If you don't specify a subnet ID, we choose a default subnet from your default VPC for you.
            If you don't have a default VPC, you must specify a subnet ID in the request.
        - [EC2-Classic] If don't specify an Availability Zone, we choose one for you.
            Some instance types must be launched into a VPC.
            If you do not have a default VPC, or if you do not specify a subnet ID, the request fails.
            For more information, see Instance types available only in a VPC.
        - [EC2-VPC] All instances have a network interface with a primary private IPv4 address.
            If you don't specify this address, we choose one from the IPv4 range of your subnet.
            Not all instance types support IPv6 addresses.
            For more information, see Instance types.
        - If you don't specify a security group ID, we use the default security group.
            For more information, see Security groups.
        - If any of the AMIs have a product code attached for which the user has not subscribed, the request fails.
        - You can create a launch template, which is a resource that contains the parameters to launch an instance.
            You can specify the launch template instead of specifying the launch parameters.
            An instance is ready for you to use when it's in the running state.
        - You can tag instances and EBS volumes during launch, after launch, or both.
        - Linux instances have access to the public key of the key pair at boot.
            You can use this key to provide secure access to the instance.
            Amazon EC2 public images use this feature to provide secure access without passwords.
            For more information, see Key pairs.

    Args:
        hub:
        ctx:
        name(Text): An Idem name of the resource.
        resource_id(Text): AWS Ec2 Instance ID
        tags(Dict, optional): The tags to apply to the resource. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        image_id(Text): The ID of an AMI
        instance_type(Text): The instance type to use for this instance on creation
        block_device_mappings((List[Dict[Text, Text]]):, optional): The block device mapping, which defines the EBS volumes and instance store volumes to attach to
            the instance at launch. For more information, see Block device mappings in the Amazon EC2 User
            Guide. Defaults to None.
            * (DeviceName, optional): The device name (for example, /dev/sdh or xvdh).
            * (VirtualName, optional): The virtual device name (ephemeralN). Instance store volumes are numbered starting from 0. An
                instance type with 2 available instance store volumes can specify mappings for ephemeral0 and
                ephemeral1. The number of available instance store volumes depends on the instance type. After
                you connect to the instance, you must mount the volume. NVMe instance store volumes are
                automatically enumerated and assigned a device name. Including them in your block device mapping
                has no effect. Constraints: For M3 instances, you must specify instance store volumes in the
                block device mapping for the instance. When you launch an M3 instance, we ignore any instance
                store volumes specified in the block device mapping for the AMI.
            * (Ebs, optional): Parameters used to automatically set up EBS volumes when the instance is launched.
            * (NoDevice, optional): To omit the device from the block device mapping, specify an empty string. When this property is
                specified, the device is removed from the block device mapping regardless of the assigned value.
        ebs_optimized(bool): Indicates whether the instance is optimized ofr Amazon EBS I/O.
        kernel_id(Text): The kernel associated with this instance, if applicable.
        placement(Dict[Text, Any]): The location where the instance launched, if applicable
        subnet_id(Text): The ID of the subnet in which the instance is running
        network_interfaces(List[Dict[Text, Any]], optional): The network interfaces to associate with the instance. If you specify a network interface, you
            must specify any security groups and subnets as part of the network interface. Defaults to None.
            * (AssociatePublicIpAddress, optional): Indicates whether to assign a public IPv4 address to an instance you launch in a VPC. The public
                IP address can only be assigned to a network interface for eth0, and can only be assigned to a
                new network interface, not an existing one. You cannot specify more than one network interface
                in the request. If launching into a default subnet, the default value is true.
            * (DeleteOnTermination, optional): If set to true, the interface is deleted when the instance is terminated. You can specify true
                only if creating a new network interface when launching an instance.
            * (Description, optional): The description of the network interface. Applies only if creating a network interface when
                launching an instance.
            * (DeviceIndex, optional): The position of the network interface in the attachment order. A primary network interface has a
                device index of 0. If you specify a network interface when launching an instance, you must
                specify the device index.
            * (Groups, optional): The IDs of the security groups for the network interface. Applies only if creating a network
                interface when launching an instance.
            * (Ipv6AddressCount, optional): A number of IPv6 addresses to assign to the network interface. Amazon EC2 chooses the IPv6
                addresses from the range of the subnet. You cannot specify this option and the option to assign
                specific IPv6 addresses in the same request. You can specify this option if you've specified a
                minimum number of instances to launch.
            * (Ipv6Addresses, optional): One or more IPv6 addresses to assign to the network interface. You cannot specify this option
                and the option to assign a number of IPv6 addresses in the same request. You cannot specify this
                option if you've specified a minimum number of instances to launch.
            * (NetworkInterfaceId, optional): The ID of the network interface. If you are creating a Spot Fleet, omit this parameter because
                you can’t specify a network interface ID in a launch specification.
            * (PrivateIpAddress, optional): The private IPv4 address of the network interface. Applies only if creating a network interface
                when launching an instance. You cannot specify this option if you're launching more than one
                instance in a RunInstances request.
            * (PrivateIpAddresses, optional): One or more private IPv4 addresses to assign to the network interface. Only one private IPv4
                address can be designated as primary. You cannot specify this option if you're launching more
                than one instance in a RunInstances request.
            * (SecondaryPrivateIpAddressCount, optional): The number of secondary private IPv4 addresses. You can't specify this option and specify more
                than one private IP address using the private IP addresses option. You cannot specify this
                option if you're launching more than one instance in a RunInstances request.
            * (SubnetId, optional): The ID of the subnet associated with the network interface. Applies only if creating a network
                interface when launching an instance.
            * (AssociateCarrierIpAddress, optional): Indicates whether to assign a carrier IP address to the network interface. You can only assign a
                carrier IP address to a network interface that is in a subnet in a Wavelength Zone. For more
                information about carrier IP addresses, see Carrier IP addresses in the Amazon Web Services
                Wavelength Developer Guide.
            * (InterfaceType, optional): The type of network interface. Valid values: interface | efa
            * (NetworkCardIndex, optional): The index of the network card. Some instance types support multiple network cards. The primary
                network interface must be assigned to network card index 0. The default is network card index 0.
                If you are using RequestSpotInstances to create Spot Instances, omit this parameter because you
                can’t specify the network card index when using this API. To specify the network card index, use
                RunInstances.
            * (Ipv4Prefixes, optional): One or more IPv4 delegated prefixes to be assigned to the network interface. You cannot use this
                option if you use the Ipv4PrefixCount option.
            * (Ipv4PrefixCount, optional): The number of IPv4 delegated prefixes to be automatically assigned to the network interface. You
                cannot use this option if you use the Ipv4Prefix option.
            * (Ipv6Prefixes, optional): One or more IPv6 delegated prefixes to be assigned to the network interface. You cannot use this
                option if you use the Ipv6PrefixCount option.
            * (Ipv6PrefixCount, optional): The number of IPv6 delegated prefixes to be automatically assigned to the network interface. You
                cannot use this option if you use the Ipv6Prefix option.
        monitoring_enabled(bool): Indicates whether detailed monitoring is enabled.
        root_device_name(Text): The device name of the root device (for example, /dev/sda1).
        product_codes(List[Dict[Text, Text]]: The product codes attached to the instance, if applicable.
        source_dest_check(bool): Indicates whether source/destination checking is enabled
        running(bool): Indicates whether the instance should be in the "running" state
        private_ip_address(Text): The Ipv4 address of the network interface within the subnet.
        reservation_id(Text): The ID of the reservation
        owner_id(Text): The ID of the AWS account that owns the reservation.
        user_data(Text): The user data for the instance
        disable_api_termination(bool): Indicates that an instance cannot be terminated using the Amazon Ec2 console, command line tool, or API.
        instance_initiated_shutdown_behavior(Text): Indicates whether an instance stops or terminates when you initiate shutdown from the instance (using the operating system command for system shutdown)
        enclave_options_enabled(bool): Indicates whether the instance is enabled for AWS Nitro Enclaves.
        ram_disk_id(Text): The ID of the RAM disk, if applicable.
        elastic_gpu_specifications(List, optional): An elastic GPU to associate with the instance. An Elastic GPU is a GPU resource that you can
            attach to your Windows instance to accelerate the graphics performance of your applications. For
            more information, see Amazon EC2 Elastic GPUs in the Amazon EC2 User Guide. Defaults to None.
            * (Type): The type of Elastic Graphics accelerator. For more information about the values to specify for
                Type, see Elastic Graphics Basics, specifically the Elastic Graphics accelerator column, in the
                Amazon Elastic Compute Cloud User Guide for Windows Instances.
        elastic_inference_accelerators(List, optional): An elastic inference accelerator to associate with the instance. Elastic inference accelerators
            are a resource you can attach to your Amazon EC2 instances to accelerate your Deep Learning (DL)
            inference workloads. You cannot specify accelerators from different generations in the same
            request. Defaults to None.
            * (Type):  The type of elastic inference accelerator. The possible values are eia1.medium, eia1.large,
                eia1.xlarge, eia2.medium, eia2.large, and eia2.xlarge.
            * (Count, optional):  The number of elastic inference accelerators to attach to the instance.  Default: 1
        iam_instance_profile(Dict[Text, Text]): The IAM instance profile
        key_name(Text): The name of the keypair
        instance_market_options(Dict[Text, Any]): The market (purchasing) option for the instance
        credit_specification(Dict[Text, Any]): The credit option for CPU usage of the instance.
        cpu_options(Dict[Text, Any]): The market (purchasing) option for the instance
        capacity_reservation_specification(Dict[Text, Text]): Information about the Capacity Reservation targeting option
        license_specifications(List[Dict[Text, Text]], optional): The license configurations. Defaults to None.
            * (LicenseConfigurationArn, optional): The Amazon Resource Name (ARN) of the license configuration.
        hibernation_enabled(bool): Indicates whether the instance is configured for hibernation.
        metadata_options(Dict[Text, Any]): The metadata options for the instance
        instance_requirements(Dict[Text, Any]): The attributes for the instance type.
            When you specify instance attributes, Amazon EC2 will identify instance types with these attributes.
        private_dns_name_options(Dict[Text, Any): The options for the instance hostname.
        maintenance_options(Dict[Text, Any]): The maintenance options for the instance.
        sriov_net_support(Text): Specifies whether enhanced networking with the Intel 82599 Virtual Function interface is enabled
        bootstrap(List[Dict[Text, Any]]): BootText options for provisioning an instance with "heist"



    Returns:
        Dict[Text, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.ec2.instance.present:
                - name: value
                - image_id: my_ami_id-0000000000000000
                - tags:
                    my_tag_key_1: my_tag_value_1
                    my_tag_key_2: my_tag_value_2
                - bootstrap:
                    - heist_manager: salt.minion
                      artifact_version: 1234

    """
    # Get all the parameters passed to this function as a single dictionary
    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "name", "resource_id", "kwargs")
    }

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    for key in kwargs:
        result["comment"] += [f"Not explicitly supported keyword argument: '{key}'"]

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # Get the resource_id from the idempotency token
    if not resource_id:
        ret = await hub.tool.aws.ec2.instance.state.get(ctx, client_token=name)
        if ret:
            resource_id = ret["resource_id"]

    if resource_id:
        # Assume that the instance already exists since we have a resource_id
        resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", resource_id)
        current_state = result["old_state"] = await hub.tool.aws.ec2.instance.state.get(
            ctx, instance_id=resource_id
        )

        if not result["old_state"]:
            result["comment"] += [
                f"Could not find instance for '{name}' with existing id '{resource_id}'"
            ]
            return result

        result["comment"] += [f"Instance '{name}' already exists"]
    else:
        # Create the instance if it doesn't already exist
        if ctx.test:
            result["new_state"] = hub.tool.aws.ec2.instance.state.test(**desired_state)
            result["comment"] += [f"Would create aws.ec2.instance '{name}'"]
            return result

        # Create a brand new instance with minimal arguments, it will be updated after creation
        create_ret = await hub.exec.boto3.client.ec2.run_instances(
            ctx,
            ClientToken=name,
            MaxCount=1,
            MinCount=1,
            # ramdisk and kernel can only be specified when an instance is stopped or on creation
            KernelId=kernel_id,
            RamDiskId=ram_disk_id,
            # We only need to add options that can ONLY be specified on creation or that are best specified at creation
            # If parameters have to be processed it's better to let them be handled by "update" plugins after creation
            ImageId=image_id,
            Placement=placement,
            SubnetId=subnet_id,
            UserData=user_data,
            EbsOptimized=ebs_optimized,
            PrivateIpAddress=private_ip_address,
            ElasticGpuSpecification=elastic_gpu_specifications,
            ElasticInferenceAccelerators=elastic_inference_accelerators,
            CreditSpecification=credit_specification,
            CpuOptions=cpu_options,
            CapacityReservationSpecification=capacity_reservation_specification,
            LicenseSpecifications=license_specifications,
            MetadataOptions=metadata_options,
            PrivateDnsNameOptions=private_dns_name_options,
            MaintenanceOptions=maintenance_options,
        )

        result["result"] &= create_ret.result
        if not create_ret:
            result["comment"] += [create_ret.comment]
            return result

        result["comment"] += [f"Created '{name}'"]
        resource_id = create_ret.ret["Instances"][0]["InstanceId"]
        # This makes sure the created VPC is saved to esm regardless if the subsequent update call fails or not.
        result["new_state"] = dict(name=name, resource_id=resource_id, **desired_state)
        result["force_save"] = True

        resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", resource_id)

        hub.log.debug(f"Waiting for instance '{name}' to be created")

        present_ret = await hub.tool.aws.ec2.instance.state.convert_to_present(
            ctx, {"Reservations": [create_ret.ret]}
        )
        current_state = present_ret[resource_id]

    if not current_state:
        result["comment"] += [
            f"Unable to get the current_state, instance may still be undergoing creation: '{name}'"
        ]
        result["result"] = False
        return result

    changes_made = False
    # Compare the kwargs of this function to the presentized attributes of the instance
    for attribute, new_value in desired_state.items():
        if new_value is None:
            # No value has been explicitly given, leave this parameter alone
            continue

        old_value = current_state.get(attribute)
        if old_value != new_value:
            changes_made = True
            # There is a single file dedicated to updating each attribute of an ec2 instance.
            # Organization is key for managing such a large resource.
            # This "present" function should remain mostly the same, don't bloat it!
            # Add an attribute-specific file in idem_aws/tool/aws/ec2/instance/update to manage specific parameters.
            # Be sure to follow the contracts in idem_aws/tool/aws/ec2/instance/update/contracts
            if attribute not in hub.tool.aws.ec2.instance.update._loaded:
                result["comment"] += [
                    f"Modifying aws.ec2.instance attribute '{attribute}' is not yet supported"
                ]
                continue
            if ctx.test:
                result["comment"] += [
                    f"Would update aws.ec2.instance '{name}': {attribute}"
                ]
                continue
            # Call the appropriate tool to update each parameter that needs updating
            result["result"] &= await hub.tool.aws.ec2.instance.update[attribute].apply(
                ctx,
                resource,
                old_value=old_value,
                new_value=new_value,
                # This list is stored in memory
                # modifying this value in "update.apply" functions will update it in the "result" dictionary
                comments=result["comment"],
            )
            if not result["result"]:
                result["comment"] += [
                    f"Unable to update aws.ec2.instance attribute: {attribute}"
                ]
                return result

    # Get the final state of the resource
    if changes_made:
        if ctx.test:
            result["new_state"] = hub.tool.aws.ec2.instance.state.test(**desired_state)
        else:
            result["new_state"] = await hub.tool.aws.ec2.instance.state.get(
                ctx, instance_id=resource_id
            )
    else:
        result["new_state"] = current_state

    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, **kwargs
) -> Dict[str, Any]:
    """
    Shuts down the specified instance.
    Terminated instances remain visible after termination (for approximately one hour).

    Args:
        hub:
        ctx:
        name(Text): The name of the state.
        resource_id(Text): An instance id

    Returns:
        Dict[Text, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.instance.absent:
                - name: value
    """
    result = dict(
        comment=[], old_state=ctx.old_state, new_state=None, name=name, result=True
    )

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # Get the resource_id from the idempotency token
    if not resource_id:
        ret = await hub.tool.aws.ec2.instance.state.get(ctx, client_token=name)
        if ret:
            resource_id = ret["resource_id"]

    # If there still is no resource_id, the instance is gone
    if not resource_id:
        result["comment"] += [f"'{name}' already terminated"]
        return result

    if ctx.test:
        result["comment"] += [f"Would terminate aws.ec2.instance '{name}'"]
        return result

    ret = await hub.exec.boto3.client.ec2.terminate_instances(
        ctx, InstanceIds=[resource_id]
    )
    result["result"] &= ret["result"]
    if not result["result"]:
        result["comment"].append(ret["comment"])
        return result
    result["comment"] += [
        f"Terminated instance '{name}', it will still be visible for about 60 minutes"
    ]

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[Text, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.instance
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_instances(ctx)

    if not ret:
        hub.log.debug(f"Could not describe Instances: {ret.comment}")
        return {}

    instances = await hub.tool.aws.ec2.instance.state.convert_to_present(ctx, ret.ret)

    for instance_id, present_state in instances.items():
        result[instance_id] = {
            "aws.ec2.instance.present": [{k: v} for k, v in present_state.items()]
        }

    return result


async def search(hub, ctx, name: str, filters: List = None, resource_id: str = None):
    """
    Use an un-managed Instance as a data-source. Supply one of the inputs as the filter.

    Args:
        hub:
        ctx:
        name(Text): The name of the Idem state.
        resource_id(Text, optional): AWS Ec2 Instance id to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key.
        A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_instances

    Request Syntax:
        [Idem-state-name]:
          aws.ec2.instance.search:
          - resource_id: 'Text'
          - filters:
            - name: 'string'
              values: 'list'
            - name: 'string'
              values: 'list'

        Examples:

            my-unmanaged-instance:
              aws.ec2.instance.search:
                - resource_id: value
    """
    result = dict(
        comment=[], old_state=ctx.old_state, new_state=None, name=name, result=True
    )

    # Get the resource_id from ESM if this resource used to be managed by "present"
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # Perform validation on the parameters
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    result["result"] &= syntax_validation["result"]
    if not result["result"]:
        result["comment"].append(syntax_validation["comment"])
        return result

    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )

    # Get all instances that match the given filters
    ret = await hub.exec.boto3.client.ec2.describe_instances(
        ctx,
        Filters=boto3_filter,
        InstanceIds=[resource_id] if resource_id else None,
    )
    result["result"] &= ret.result
    if not ret:
        result["comment"].append(ret.comment)
        return result

    # Convert the described instances to the present state format
    instances = await hub.tool.aws.ec2.instance.state.convert_to_present(ctx, ret.ret)

    # Check for null results
    if not instances:
        result["result"] = False
        result["comment"] += [
            f"Unable to find an aws.ec2.instance for '{name}' that matched the given filters"
        ]
        return result

    # Get the first instance from the results
    instance_id = next(iter(instances.keys()))

    # Add a comment if there were multiple results
    if len(instances) > 1:
        result["comment"] += [
            f"More than one aws.ec2.instance resource was found. Use resource '{instance_id}'"
        ]

    # Return both old_state and new_state together
    result["new_state"] = instances[instance_id]

    return result



import botocore.client
import botocore.docs.docstring
import botocore.exceptions
import botocore.model

__func_alias__ = {"type_": "type"}


def parse(hub, param: "botocore.model.Shape", required: bool):
    docstring = hub.tool.format.html.parse(param.documentation)
    ret = {
        "required": required,
        "default": None,
        "target_type": "mapping",
        "target": "kwargs",
        "param_type": hub.pop_create.aws.param.type(param),
        "doc": "\n            ".join(
            hub.tool.format.wrap.wrap(
                docstring,
                width=96,
            )
        ),
    }
    member = None
    if isinstance(param, botocore.model.StructureShape):
        member = param
    elif param.type_name == "list" and isinstance(
        param.member, botocore.model.StructureShape
    ):
        member = param.member
    elif param.type_name == "map" and isinstance(
        param.value, botocore.model.StructureShape
    ):
        member = param.value

    if member:
        ret["member"] = {
            "name": param.name,
            "params": {
                k: parse(hub, v, k in member.required_members)
                for k, v in member.members.items()
            },
        }
    return ret


def type_(hub, param: "botocore.model.Shape"):
    if param.type_name == "string":
        return "str"
    elif param.type_name == "map":
        return f"Dict[{hub.pop_create.aws.param.type(param.key)}, {hub.pop_create.aws.param.type(param.value)}]"
    elif param.type_name == "structure":
        return "Dict"
    elif param.type_name == "list":
        return f"List[{hub.pop_create.aws.param.type(param.member)}]"
    elif param.type_name == "boolean":
        return "bool"
    elif param.type_name in ("integer", "long"):
        return "int"
    elif param.type_name in ("float", "double"):
        return "float"
    elif param.type_name == "timestamp":
        return "Text"
    elif param.type_name == "blob":
        return "ByteString"
    else:
        raise NameError(param.type_name)



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS Cloudwatch to present input format.
"""


async def convert_raw_log_group_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("logGroupName")
    resource_parameters = OrderedDict(
        {"kmsKeyId": "kms_key_id", "arn": "arn", "retentionInDays": "retention_in_days"}
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if resource_id:
        # list_tags_log_group always returns true even if there is no tag
        tags = await hub.exec.boto3.client.logs.list_tags_log_group(
            ctx, logGroupName=resource_id
        )
        if tags["result"]:
            if tags["ret"]["tags"]:
                resource_translated["tags"] = dict(tags["ret"]["tags"])
    return resource_translated


async def convert_raw_metric_alarm_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    r"""

    Args:
        hub:
        ctx:
        raw_resource: old state of metric alarm
        idem_resource_name: cloudwatch alarm name

    Returns:
        Dict[str, Any]
    """

    resource_parameters = OrderedDict(
        {
            "AlarmDescription": "alarm_description",
            "ActionsEnabled": "actions_enabled",
            "OKActions": "ok_actions",
            "AlarmActions": "alarm_actions",
            "InsufficientDataActions": "insufficient_data_actions",
            "MetricName": "metric_name",
            "Namespace": "namespace",
            "Statistic": "statistic",
            "ExtendedStatistic": "extended_statistic",
            "Dimensions": "dimensions",
            "Period": "period",
            "Unit": "unit",
            "EvaluationPeriods": "evaluation_periods",
            "DatapointsToAlarm": "datapoints_to_alarm",
            "Threshold": "threshold",
            "ComparisonOperator": "comparison_operator",
            "TreatMissingData": "treat_missing_data",
            "EvaluateLowSampleCountPercentile": "evaluate_low_sample_count_percentile",
            "Metrics": "metrics",
            "ThresholdMetricId": "threshold_metric_id",
            "Tags": "tags",
        }
    )
    resource_id = raw_resource.get("AlarmName")
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
    }
    result = dict(comment=(), result=True, ret=None)
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource[parameter_raw]

    alarm_arn = raw_resource.get("AlarmArn")
    ret_tag = await hub.exec.boto3.client.cloudwatch.list_tags_for_resource(
        ctx, ResourceARN=alarm_arn
    )
    result["result"] = ret_tag["result"]
    if not result["result"]:
        result["comment"] = result["comment"] + ret_tag["comment"]

    if result["result"] and ret_tag["ret"].get("Tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            ret_tag["ret"]["Tags"]
        )
    result["ret"] = resource_translated

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS Config rule to present input format.
"""


async def convert_raw_config_rule_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    # ConfigRuleName is the unique identifier for Config so it is set as resource_id
    resource_id = raw_resource.get("ConfigRuleName")
    config_rule_arn = raw_resource.get("ConfigRuleArn")
    tag_ret = await hub.exec.boto3.client.config.list_tags_for_resource(
        ctx, ResourceArn=config_rule_arn
    )

    resource_parameters = OrderedDict(
        {
            "ConfigRuleName": "config_rule_name",
            "ConfigRuleArn": "config_rule_arn",
            "ConfigRuleId": "config_rule_id",
            "Scope": "scope",
            "Source": "source",
            "ConfigRuleState": "config_rule_state",
            "MaximumExecutionFrequency": "max_execution_frequency",
            "InputParameters": "input_parameters",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    if tag_ret:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            tag_ret["ret"]["Tags"]
        )
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated


async def convert_raw_config_recorder_to_present(
    hub, ctx, raw_resource: Dict[str, Any], recording: bool
) -> Dict[str, Any]:
    resource_translated = {
        "name": raw_resource.get("name"),
        "resource_id": raw_resource.get("name"),
        "role_arn": raw_resource.get("roleARN"),
        "recording": recording,
        "recording_group": raw_resource.get("recordingGroup"),
    }
    return resource_translated


async def convert_raw_config_delivery_channel_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    # name is the unique identifier for Config so it is set as resource_id
    resource_id = raw_resource.get("name")

    resource_parameters = OrderedDict(
        {
            "name": "name",
            "s3BucketName": "s3_bucket_name",
            "s3KeyPrefix": "s3_key_prefix",
            "s3KmsKeyArn": "s3_kms_key_arn",
            "snsTopicARN": "sns_topic_arn",
            "configSnapshotDeliveryProperties": "config_snapshot_delivery_properties",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated



import datetime
import io
import os
import re
import sys
import tempfile
from typing import Any
from typing import Dict
from typing import Tuple

import msgpack
from cryptography.fernet import Fernet
from dict_tools import data

try:
    import aws_google_auth

    HAS_GOOGLE_AUTH = True
except ImportError as e:
    HAS_GOOGLE_AUTH = False, str(e)


def __virtual__(hub):
    return HAS_GOOGLE_AUTH


def __init__(hub):
    # Initialize a directory where credentials can be stored
    hub.acct.aws.gsuite.TMP = os.path.join(
        tempfile.gettempdir(), "idem", "aws_google_acct"
    )
    os.makedirs(hub.acct.aws.gsuite.TMP, exist_ok=True)


def parse_opts(
    hub,
    role_arn: str,
    username: str = None,
    password: str = None,
    duration: int = None,
    idp_id: str = None,
    region: str = None,
    sp_id: str = None,
    resolve_aliases: bool = None,
    account: str = None,
    keyring: str = None,
    saml_assertion: str = None,
    saml_cache: bool = True,
    **ctx,
):
    args = data.NamespaceDict(
        {
            "auto_duration": False,
            "duration": duration,
            "idp_id": idp_id,
            "region": region,
            "role_arn": role_arn,
            "sp_id": sp_id,
            "resolve_aliases": resolve_aliases,
            "username": username,
            "account": account,
            "keyring": keyring,
            "saml_assertion": saml_assertion,
            "saml_cache": saml_cache,
            "ask_role": False,
            "disable_u2f": False,
            "profile": None,
            "bg_response": True,
            "save_failure_html": False,
            "print_creds": True,
            "quiet": True,
            "log_level": "warn",
        }
    )

    config = aws_google_auth.resolve_config(args)

    print(
        "You may be prompted to verify your identity on a secondary 2FA device before"
        " this can continue"
    )

    out = io.StringIO()
    get_pass = aws_google_auth.util.Util.get_password

    if password is not None:
        aws_google_auth.util.Util.get_password = lambda *z: password
    sys.stdout = out
    aws_google_auth.process_auth(args, config)
    sys.stdout = sys.__stdout__
    if password is not None:
        aws_google_auth.util.Util.get_password = get_pass

    stripped_out = out.getvalue().strip().replace("\n", "")

    match = re.match(
        "(.*)export AWS_ACCESS_KEY_ID='([^']+)' AWS_SECRET_ACCESS_KEY='([^']+)'"
        " AWS_SESSION_TOKEN='([^']+)' AWS_SESSION_EXPIRATION='([^']+)'(.*)",
        stripped_out,
    )
    if match is None:
        raise ConnectionRefusedError("Could not connect to aws")

    if match.group(1):
        hub.log.info(match.group(1))
    ctx["aws_access_key_id"] = match.group(2)
    ctx["aws_secret_access_key"] = match.group(3)
    ctx["aws_session_token"] = match.group(4)
    ctx["region_name"] = config.region
    session_expiration = match.group(5)
    if match.group(6):
        hub.log.info(match.group(6))
    return ctx, session_expiration


async def load_profile(hub, enc_profile: str) -> Tuple[Dict[str, Any], str]:
    if os.path.exists(enc_profile):
        try:
            ret = await hub.crypto.init.decrypt_file(
                acct_key=hub.OPT.acct.acct_key,
                acct_file=enc_profile,
                crypto_plugin="fernet",
            )
            return ret["boto_creds"], ret["expiration"]
        except Exception as e:
            hub.log.error(e)

    return {}, ""


def dump_profile(hub, enc_profile: str, boto_creds: Dict[str, Any], expiration: str):
    hub.log.debug(f"Writing encrypted aws acct information to {enc_profile}")
    f = Fernet(hub.OPT.acct.acct_key)
    with open(enc_profile, "wb+") as wfh:
        data = {"boto_creds": boto_creds, "expiration": expiration}
        wfh.write(f.encrypt(msgpack.dumps(data)))


async def gather(hub, profiles) -> Dict[str, Any]:
    """
    Generate AWS credentials from Google Auth

    Example:
    .. code-block:: yaml

        aws.gsuite:
          profile_name:
            username: example@gmail.com
            password: please_don't_use_this
            role_arn: arn:aws:iam::999999999:role/xacct/developer
            idp_id: XXXXXXXXX
            sp_id: 999999999999
            region: us-east-1
            duration: 36000
            account: developer
    """
    sub_profiles = data.NamespaceDict()
    for profile, ctx in profiles.get("aws.gsuite", {}).items():
        enc_profile = os.path.join(hub.acct.aws.gsuite.TMP, f"{profile}.fernet")
        # Try to load existing account information
        boto_creds, expiration = await hub.acct.aws.gsuite.load_profile(enc_profile)
        try:
            expiration = datetime.datetime.strptime(expiration, "%Y-%m-%dT%H:%M:%S%z")
        except ValueError:
            expiration = None

        # Recreate the session if it expired or doesn't exist
        if not (
            boto_creds
            or (expiration and expiration < datetime.datetime.now(expiration.tzinfo))
        ):
            boto_creds, expiration = hub.acct.aws.gsuite.parse_opts(**ctx)
            # Save the account information to an encrypted file
            hub.acct.aws.gsuite.dump_profile(enc_profile, boto_creds, expiration)
        else:
            hub.log.debug(
                f"Found existing google-aws-auth credentials in {enc_profile}"
            )

        # Strip any args that were used for authentication
        sub_profiles[profile] = boto_creds

    return sub_profiles



from typing import Any
from typing import Dict


async def get_updated_payload_delivery_channel(
    hub,
    ctx,
    resource_id: str,
    before: Dict[str, Any],
    s3_bucket_name: str,
    s3_key_prefix: str,
    s3_kms_key_arn: str,
    sns_topic_arn: str,
    config_snapshot_delivery_properties: Dict[str, Any],
):
    """
    Get Updated payload of delivery channel of AWS Config

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        resource_id: AWS Delivery channel name
        before: Contains current configuration for the resource
        s3_bucket_name: String. Specified S3 Bucket name for delivery channel
        s3_key_prefix: String. The prefix for the specified Amazon S3 bucket.
        s3_kms_key_arn: String. ARN of the Key Management Service (KMS ) KMS key (KMS key) used to encrypt objects
        sns_topic_arn: String. Amazon SNS topic to which Config sends notifications about configuration changes.
        config_snapshot_delivery_properties: Dict.
            delivery_frequency: String. The frequency with which Config delivers configuration snapshots.

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret=None)

    update_payload = {}
    delivery_props = {}

    if s3_bucket_name and before.get("s3_bucket_name") != s3_bucket_name:
        update_payload["s3BucketName"] = s3_bucket_name

    if s3_key_prefix and before.get("s3_key_prefix") != s3_key_prefix:
        update_payload["s3KeyPrefix"] = s3_key_prefix

    if s3_kms_key_arn and before.get("s3_kms_key_arn") != s3_kms_key_arn:
        update_payload["s3KmsKeyArn"] = s3_kms_key_arn

    if sns_topic_arn and before.get("sns_topic_arn") != sns_topic_arn:
        update_payload["snsTopicARN"] = sns_topic_arn

    if (
        config_snapshot_delivery_properties
        and not before.get("configSnapshotDeliveryProperties")
    ) or (
        config_snapshot_delivery_properties
        and before.get("configSnapshotDeliveryProperties")
        and before.get("configSnapshotDeliveryProperties").get("deliveryFrequency")
        != config_snapshot_delivery_properties.get("delivery_frequency")
    ):
        delivery_props["deliveryFrequency"] = config_snapshot_delivery_properties[
            "delivery_frequency"
        ]
        update_payload["configSnapshotDeliveryProperties"] = delivery_props

    update_payload["name"] = resource_id

    result["ret"] = update_payload
    result["result"] = True
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict


async def update_metric(
    hub,
    ctx,
    alarm_name: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
):
    """
    Update tags of AWS cloudwatch metric alarm
    Args:
        hub:
        ctx:
        alarm_name: alarm name
        raw_resource: old state of metric alarm
        resource_parameters: list of new params

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """

    parameters = OrderedDict(
        {
            "AlarmDescription": "alarm_description",
            "ActionsEnabled": "actions_enabled",
            "OKActions": "ok_actions",
            "AlarmActions": "alarm_actions",
            "InsufficientDataActions": "insufficient_data_actions",
            "MetricName": "metric_name",
            "Namespace": "namespace",
            "Statistic": "statistic",
            "ExtendedStatistic": "extended_statistic",
            "Dimensions": "dimensions",
            "Period": "period",
            "Unit": "unit",
            "EvaluationPeriods": "evaluation_periods",
            "DatapointsToAlarm": "datapoints_to_alarm",
            "Threshold": "threshold",
            "ComparisonOperator": "comparison_operator",
            "TreatMissingData": "treat_missing_data",
            "EvaluateLowSampleCountPercentile": "evaluate_low_sample_count_percentile",
            "Metrics": "metrics",
            "ThresholdMetricId": "threshold_metric_id",
        }
    )
    resource_update = {}
    result = dict(comment=(), result=True, ret=None)
    resource_parameters.pop("Tags", None)
    for key, value in resource_parameters.items():
        if value is None or value == raw_resource[key]:
            continue
        resource_update[key] = resource_parameters[key]

    if resource_update:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.cloudwatch.put_metric_alarm(
                ctx, **resource_parameters
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result["comment"] + (f"Updated '{alarm_name}'",)

        result["ret"] = {}
        for parameter_raw, parameter_present in parameters.items():
            if parameter_raw in resource_update:
                result["ret"][parameter_present] = resource_update[parameter_raw]

    return result


async def update_metric_tags(
    hub,
    ctx,
    alarm_arn,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """
    Update tags of AWS CloudWatch resources
    Args:
        alarm_arn:
        hub:
        ctx:
        old_tags (Dict): Dict in the format of {tag-key: tag-value}
        new_tags (Dict): Dict in the format of {tag-key: tag-value}

    Returns:
        {"result": True|False, "comment": "A message", "ret": dict of updated tags}

    """
    tags_to_add = {}
    tags_to_remove = {}

    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.cloudwatch.untag_resource(
                ctx, ResourceARN=alarm_arn, TagKeys=list(tags_to_remove)
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.cloudwatch.tag_resource(
                ctx,
                ResourceARN=alarm_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = new_tags
    result["comment"] = (
        f"Updated tags: Added [{tags_to_add}] Removed [{tags_to_remove}]",
    )
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_vpc
hub.exec.boto3.client.ec2.delete_vpc
hub.exec.boto3.client.ec2.describe_vpcs
resource = hub.tool.boto3.resource.create(ctx, "ec2", "Vpc", name)
hub.tool.boto3.resource.exec(resource, associate_dhcp_options, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_classic_link_instance, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_internet_gateway, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_network_acl, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_route_table, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_security_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_subnet, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, describe_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_classic_link_instance, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_internet_gateway, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, disable_classic_link, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, enable_classic_link, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, modify_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, request_vpc_peering_connection, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": [
            "aws.ec2.subnet.absent",
            "aws.ec2.security_group.absent",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    cidr_block_association_set: List[
        make_dataclass(
            "CidrBlockAssociationSet",
            [
                ("CidrBlock", str, field(default=None)),
                ("Ipv4IpamPoolId", str, field(default=None)),
                ("Ipv4NetmaskLength", int, field(default=0)),
            ],
        )
    ] = None,
    ipv6_cidr_block_association_set: List[
        make_dataclass(
            "Ipv6CidrBlockAssociationSet",
            [
                ("Ipv6CidrBlock", str, field(default=None)),
                ("Ipv6IpamPoolId", str, field(default=None)),
                ("Ipv6NetmaskLength", int, field(default=0)),
                ("Ipv6CidrBlockNetworkBorderGroup", str, field(default=None)),
                ("AmazonProvidedIpv6CidrBlock", bool, field(default=False)),
            ],
        )
    ] = None,
    instance_tenancy: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    enable_dns_hostnames: bool = None,
    enable_dns_support: bool = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a VPC with the specified IPv4 CIDR block. The smallest VPC you can create uses a /28 netmask (16 IPv4
    addresses), and the largest uses a /16 netmask (65,536 IPv4 addresses). For more information about how large to
    make your VPC, see Your VPC and subnets in the Amazon Virtual Private Cloud User Guide. You can optionally
    request an IPv6 CIDR block for the VPC. You can request an Amazon-provided IPv6 CIDR block from Amazon's pool of
    IPv6 addresses, or an IPv6 CIDR block from an IPv6 address pool that you provisioned through bring your own IP
    addresses (BYOIP). By default, each instance you launch in the VPC has the default DHCP options, which include
    only a default DNS server that we provide (AmazonProvidedDNS). For more information, see DHCP options sets in
    the Amazon Virtual Private Cloud User Guide. You can specify the instance tenancy value for the VPC when you
    create it. You can't change this value for the VPC after you create it. For more information, see Dedicated
    Instances in the Amazon Elastic Compute Cloud User Guide.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): AWS VPC ID
        cidr_block_association_set(List, optional): Information about the IPv4 CIDR blocks associated with the VPC.
            Defaults to None.
            * CidrBlock (string) -- An IPv4 CIDR block to associate with the VPC.
            * Ipv4IpamPoolId (string) -- Associate a CIDR allocated from an IPv4 IPAM pool to a VPC.
            * Ipv4NetmaskLength (integer) -- The netmask length of the IPv4 CIDR you would like to associate from an Amazon VPC IP Address Manager (IPAM) pool.
        ipv6_cidr_block_association_set(List, optional): Information about the IPv6 CIDR blocks associated with the VPC.
            Defaults to None.
            * Ipv6CidrBlock (string) -- An IPv6 CIDR block from the IPv6 address pool. You must also specify Ipv6Pool in the request.
            * Ipv6IpamPoolId (string) -- Associates a CIDR allocated from an IPv6 IPAM pool to a VPC.
            * Ipv6NetmaskLength (integer) -- The netmask length of the IPv6 CIDR you would like to associate from an
                Amazon VPC IP Address Manager (IPAM) pool.
            * Ipv6CidrBlockNetworkBorderGroup (string) -- The name of the location from which we advertise the IPV6 CIDR
                block. Use this parameter to limit the CIDR block to this location. You must set AmazonProvidedIpv6CidrBlock
                to true to use this parameter. You can have one IPv6 CIDR block association per network border group.
            * AmazonProvidedIpv6CidrBlock (boolean) -- Requests an Amazon-provided IPv6 CIDR block with a /56 prefix
                length for the VPC. You cannot specify the range of IPv6 addresses, or the size of the CIDR block.
        instance_tenancy(Text, optional): The tenancy options for instances launched into the VPC. For default, instances are launched
            with shared tenancy by default. You can launch instances with any tenancy into a shared tenancy
            VPC. For dedicated, instances are launched as dedicated tenancy instances by default. You can
            only launch instances with a tenancy of dedicated or host into a dedicated tenancy VPC.
            Important: The host value cannot be used with this parameter. Use the default or dedicated
            values only. Default: default. Defaults to None.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the VPC.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        enable_dns_hostnames(bool, optional): Indicates whether the instances launched in the VPC get DNS hostnames. If enabled, instances
            in the VPC get DNS hostnames; otherwise, they do not. You cannot modify the DNS resolution and DNS hostnames attributes in the same request.
            Use separate requests for each attribute. You can only enable DNS hostnames if you've enabled DNS support.
        enable_dns_support(bool, optional): Indicates whether the DNS resolution is supported for the VPC. If enabled, queries to the Amazon provided
            DNS server at the 169.254.169.253 IP address, or the reserved IP address at the base of the VPC network range "plus two" succeed. If disabled,
            the Amazon provided DNS service in the VPC that resolves public DNS hostnames to IP addresses is not enabled. You cannot modify the DNS resolution
            and DNS hostnames attributes in the same request. Use separate requests for each attribute.

    Request Syntax:
        [vpc-resource-id]:
          aws.ec2.vpc.present:
          - resource_id: 'string'
          - cidr_block_association_set:
            - CidrBlock: 'string'
              Ipv4IpamPoolId: 'string'
              Ipv4NetmaskLength: 'integer'
          - ipv6_cidr_block_association_set:
            - Ipv6CidrBlock: 'string'
              Ipv6IpamPoolId: 'string'
              Ipv6NetmaskLength: 'integer'
              Ipv6CidrBlockNetworkBorderGroup: 'string'
              AmazonProvidedIpv6CidrBlock: True|False
          - instance_tenancy: 'default'|'dedicated'|'host'
          - tags:
            - Key: 'string'
              Value: 'string'
          - enable_dns_support: 'Boolean'
          - enable_dns_hostnames: 'Boolean'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            vpc-01234672f3336db8:
              aws.ec2.vpc.present:
              - cidr_block_association_set:
                - CidrBlock: 10.1.150.0/28
              - instance_tenancy: default
              - enable_dns_support: True
              - enable_dns_hostnames: False
              - tags:
                - Key: Name
                  Value: vpc-name
                - Key: vpc-tag-key-2
                  Value: vpc-tag-value-2
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if resource_id:
        before = await hub.exec.aws.ec2.vpc.get(ctx, name=name, resource_id=resource_id)
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        # Update cidr blocks
        update_ret = await hub.tool.aws.ec2.vpc.update_cidr_blocks(
            ctx=ctx,
            vpc_id=result["old_state"]["resource_id"],
            old_ipv4_cidr_blocks=result["old_state"].get(
                "cidr_block_association_set", []
            ),
            old_ipv6_cidr_blocks=result["old_state"].get(
                "ipv6_cidr_block_association_set", []
            ),
            new_ipv4_cidr_blocks=cidr_block_association_set,
            new_ipv6_cidr_blocks=ipv6_cidr_block_association_set,
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            if update_ret["ret"].get("cidr_block_association_set") is not None:
                plan_state["cidr_block_association_set"] = update_ret["ret"].get(
                    "cidr_block_association_set"
                )
            if update_ret["ret"].get("ipv6_cidr_block_association_set") is not None:
                plan_state["ipv6_cidr_block_association_set"] = update_ret["ret"].get(
                    "ipv6_cidr_block_association_set"
                )
        # modify vpc attribute if the old and new attributes are not same
        if enable_dns_hostnames is not None or enable_dns_support is not None:
            update_ret = await hub.tool.aws.ec2.vpc.update_vpc_attributes(
                ctx, enable_dns_hostnames, enable_dns_support, resource_id
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                if enable_dns_support is not None:
                    plan_state["enable_dns_support"] = enable_dns_support
                if enable_dns_hostnames is not None and enable_dns_support:
                    plan_state["enable_dns_hostnames"] = enable_dns_hostnames
        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=result["old_state"].get("resource_id"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["result"])
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cidr_block_association_set": cidr_block_association_set,
                    "ipv6_cidr_block_association_set": ipv6_cidr_block_association_set,
                    "instance_tenancy": instance_tenancy,
                    "tags": tags,
                    "enable_dns_hostnames": enable_dns_hostnames,
                    "enable_dns_support": enable_dns_support,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.vpc", name=name
            )
            return result
        cidr_request_payload = {}
        # Since boto3 only allows one cidr association when creating a vpc, we use the first cidr associations
        # during vpc creation, associate the rest after creation.
        if cidr_block_association_set:
            cidr_request_payload = (
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    cidr_block_association_set[0], "ipv4"
                )
            )
            cidr_block_association_set.pop(0)
        elif ipv6_cidr_block_association_set:
            cidr_request_payload = (
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    ipv6_cidr_block_association_set[0], "ipv6"
                )
            )
            ipv6_cidr_block_association_set.pop(0)

        ret = await hub.exec.boto3.client.ec2.create_vpc(
            ctx,
            InstanceTenancy=instance_tenancy,
            TagSpecifications=[
                {
                    "ResourceType": "vpc",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags
            else None,
            **cidr_request_payload,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["Vpc"]["VpcId"]
        # This makes sure the created VPC is saved to esm regardless if the subsequent update call fails or not.
        result["new_state"] = {"name": name, "resource_id": resource_id}
        result["force_save"] = True
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.vpc", name=name
        )
        # Associate the rest cidr associations
        update_ret = await hub.tool.aws.ec2.vpc.update_cidr_blocks(
            ctx=ctx,
            vpc_id=resource_id,
            old_ipv4_cidr_blocks=[],
            old_ipv6_cidr_blocks=[],
            new_ipv4_cidr_blocks=cidr_block_association_set,
            new_ipv6_cidr_blocks=ipv6_cidr_block_association_set,
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = result["result"] and update_ret["result"]

        # modify vpc attribute if the old and new attributes are not same
        if enable_dns_hostnames is not None or enable_dns_support is not None:
            update_ret = await hub.tool.aws.ec2.vpc.update_vpc_attributes(
                ctx, enable_dns_hostnames, enable_dns_support, resource_id
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not result["old_state"]) or resource_updated:
            after = await hub.exec.aws.ec2.vpc.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified VPC. You must detach or delete all gateways and resources that are associated with the VPC
    before you can delete it. For example, you must terminate all instances running in the VPC, delete all security
    groups associated with the VPC (except the default one), delete all route tables associated with the VPC (except
    the default one), and so on.

    Args:
        name(Text): The Idem name of the VPC.
        resource_id(Text, optional): AWS VPC ID. Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            vpc-01234672f3336db8:
              aws.ec2.vpc.absent:
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.vpc", name=name
        )
        return result
    before = await hub.exec.aws.ec2.vpc.get(ctx=ctx, name=name, resource_id=resource_id)
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.vpc", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.vpc", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.ec2.delete_vpc(ctx, VpcId=resource_id)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.vpc", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_vpcs(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe VPCs {ret['comment']}")
        return {}

    for resource in ret["ret"]["Vpcs"]:
        resource_id = resource.get("VpcId")
        resource_translated = (
            await hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_to_present(
                ctx, raw_resource=resource, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.vpc.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result


async def search(
    hub, ctx, name, filters: List = None, resource_id: str = None, default: bool = None
):
    """
    Use an un-managed VPC as a data-source. Supply one of the inputs as the filter.
    """
    hub.log.warning(
        f"aws.ec2.vpc.search '{name}' state has been deprecated. Please use exec.run with aws.ec2.vpc.get instead."
    )
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    ret = await hub.exec.aws.ec2.vpc.get(
        ctx=ctx, name=name, filters=filters, resource_id=resource_id, default=default
    )
    result["result"] = ret["result"]
    result["comment"] = ret["comment"]
    if result["result"]:
        result["old_state"] = ret["ret"]
        # Populate both "old_state" and "new_state" with the same data
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.rds.create_db_subnet_group
hub.exec.boto3.client.rds.delete_db_subnet_group
hub.exec.boto3.client.rds.describe_db_subnet_groups
hub.exec.boto3.client.rds.modify_db_subnet_group
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.subnet.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    db_subnet_group_description: str,
    resource_id: str = None,
    subnets: List = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new DB subnet group. DB subnet groups must contain at least one subnet in at least two AZs in the
    Amazon Web Services Region.

    Args:
        name(Text): The Idem name of the DB subnet group.
        db_subnet_group_description(Text): DB Subnet Group description
        resource_id(Text, optional): AWS RDS DBSubnetGroupName to identify the resource
        subnets(List): The EC2 Subnet IDs for the DB subnet group
        tags(List or Dict, optional): Tags to assign to the DB subnet group. List of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                                      {tag-key: tag-value}

    Request Syntax:
        [db-subnet-group-name]:
          aws.rds.db_subnet_group.absent:
            - db_subnet_group_description: 'string'
            - subnets: 'list'
            - tags:
              - Key: 'string'
                Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            new-db-subnet-group:
                aws.rds.db_subnet_group.absent:
                  - db_subnet_group_description: "should not be blank"
                  - subnets:
                    - subnet-03e29c119d78899e1
                    - subnet-04dd5bc5aa08cbd89
                    - subnet-0754289517e31d331
                  - tags:
                    - Key: name
                      Value: new-db-subnet-group
    """

    plan_state = {}
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        before = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
            ctx, DBSubnetGroupName=resource_id
        )
    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )
    old_db_subnet_group = {}
    is_updated = False
    are_tags_updated = False
    db_subnet_group_arn = None
    old_tag = None
    if before and before["result"]:
        try:
            subnet_list = []
            old_db_subnet_group = hub.tool.aws.rds.conversion_utils.convert_raw_db_subnet_group_to_present(
                resource=before["ret"]["DBSubnetGroups"][0],
                idem_resource_name=name,
            )
            for each_subnet in old_db_subnet_group.get("subnets"):
                subnet_list.append(each_subnet)
            if set(subnets) != set(subnet_list):
                is_updated = True
            db_subnet_group_arn = old_db_subnet_group.get("db_subnet_group_arn")
            ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx, ResourceName=db_subnet_group_arn
            )
            if ret_tag["result"]:
                old_tag = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    ret_tag.get("ret").get("TagList")
                )
                old_db_subnet_group["tags"] = old_tag
            else:
                result["comment"] = ret_tag["comment"]
                result["result"] = False
                return result
            plan_state = copy.deepcopy(old_db_subnet_group)
            if ctx.get("test", False):
                plan_state["subnets"] = subnets
                plan_state["subnets"].sort()
                if db_subnet_group_description:
                    plan_state[
                        "db_subnet_group_description"
                    ] = db_subnet_group_description
            else:
                if is_updated:
                    update_ret = await hub.exec.boto3.client.rds.modify_db_subnet_group(
                        ctx,
                        DBSubnetGroupName=name,
                        DBSubnetGroupDescription=db_subnet_group_description,
                        SubnetIds=subnets,
                    )
                    if not update_ret["result"]:
                        result["comment"] = result["comment"] + update_ret["comment"]
                        result["result"] = False
                        return result
            if tags is not None and tags != old_tag:
                update_tags_ret = await hub.exec.aws.rds.tag.update_rds_tags(
                    ctx=ctx,
                    resource_arn=db_subnet_group_arn,
                    old_tags=old_tag,
                    new_tags=tags,
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = False
                    return result
                are_tags_updated = update_tags_ret["result"]
            if ctx.get("test", False):
                if are_tags_updated:
                    plan_state["tags"] = update_tags_ret["ret"]
            elif is_updated or are_tags_updated:
                result["comment"] = result["comment"] + (f"Updated '{name}'",)
            else:
                result["comment"] = result["comment"] + (f"'{name}' already exists",)
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "db_subnet_group_description": db_subnet_group_description,
                    "resource_id": resource_id,
                    "subnets": subnets,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.rds.db_subnet_group", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.rds.create_db_subnet_group(
                ctx,
                DBSubnetGroupName=name,
                DBSubnetGroupDescription=db_subnet_group_description,
                SubnetIds=subnets,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            db_subnet_group_arn = (
                ret["ret"].get("DBSubnetGroup").get("DBSubnetGroupArn")
            )
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.rds.db_subnet_group", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    try:
        if ctx.get("test", False):
            new_db_subnet_group = plan_state
        elif is_updated or not (before and before["result"]):
            after = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
                ctx, DBSubnetGroupName=name
            )
            new_db_subnet_group = hub.tool.aws.rds.conversion_utils.convert_raw_db_subnet_group_to_present(
                resource=after["ret"]["DBSubnetGroups"][0],
                idem_resource_name=name,
            )
        else:
            new_db_subnet_group = copy.deepcopy(old_db_subnet_group)

        if not ctx.get("test", False):
            if are_tags_updated or not (before and before["result"]):
                ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
                    ctx, ResourceName=db_subnet_group_arn
                )
                new_db_subnet_group[
                    "tags"
                ] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    ret_tag.get("ret").get("TagList")
                )
            else:
                new_db_subnet_group["tags"] = old_tag

        result["old_state"] = old_db_subnet_group
        result["new_state"] = new_db_subnet_group
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a DB subnet group.  The specified database subnet group must not be associated with any DB instances.

    Args:
        name(Text): The Idem name of the DB subnet group.
        resource_id(Text, optional): The AWS RDS DBSubnetGroupName to identify the resource

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_subnet_group.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_subnet_group", name=name
        )
        return result
    try:
        before = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
            ctx, DBSubnetGroupName=resource_id
        )
        if not before:
            result["comment"] = result[
                "comment"
            ] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.rds.db_subnet_group", name=name
            )
            return result
        else:
            tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx,
                ResourceName=(before["ret"]["DBSubnetGroups"][0]).get(
                    "DBSubnetGroupArn"
                ),
            )
            result[
                "old_state"
            ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_subnet_group_to_present(
                resource=before["ret"]["DBSubnetGroups"][0],
                tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    tags.get("ret").get("TagList")
                )
                if tags.get("result")
                else None,
                idem_resource_name=name,
            )
            if ctx.get("test", False):
                result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                    resource_type="aws.rds.db_subnet_group", name=name
                )
                return result
            else:
                try:
                    ret = await hub.exec.boto3.client.rds.delete_db_subnet_group(
                        ctx, DBSubnetGroupName=name
                    )
                    result["result"] = ret["result"]
                    if not result["result"]:
                        result["comment"] = ret["comment"]
                        result["result"] = False
                        return result
                    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                        resource_type="aws.rds.db_subnet_group", name=name
                    )
                except hub.tool.boto3.exception.ClientError as e:
                    result["comment"] = (f"{e.__class__.__name__}: {e}",)
                    result["result"] = False
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of DBSubnetGroup descriptions. For an overview of CIDR ranges, go to the Wikipedia Tutorial.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.rds.db_subnet_group
    """

    result = {}
    try:
        ret = await hub.exec.boto3.client.rds.describe_db_subnet_groups(ctx)
        if not ret["result"]:
            hub.log.debug(f"Could not describe db_subnet_group {ret['comment']}")
            return {}

        for db_subnet_group in ret["ret"]["DBSubnetGroups"]:
            db_subnet_group_id = db_subnet_group.get("DBSubnetGroupName")
            tags = None
            ret_tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx, ResourceName=db_subnet_group.get("DBSubnetGroupArn")
            )
            if not ret_tags:
                result["comment"] = ret_tags["comment"]
                result["result"] = False
            else:
                tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    ret_tags.get("ret").get("TagList")
                )
            resource_converted = hub.tool.aws.rds.conversion_utils.convert_raw_db_subnet_group_to_present(
                resource=db_subnet_group,
                tags=tags,
                idem_resource_name=db_subnet_group_id,
            )
            result[db_subnet_group_id] = {
                "aws.rds.db_subnet_group.present": [resource_converted]
            }
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result
